@@ -2,7 +2,7 @@
 ** wacusb.c
 **
 ** Copyright (C) 2002 - 2004 - John E. Joganic
-** Copyright (C) 2003 - 2008 - Ping Cheng
+** Copyright (C) 2003 - 2009 - Ping Cheng
 **
 ** This program is free software; you can redistribute it and/or
 ** modify it under the terms of the GNU General Public License
@@ -12,7 +12,7 @@
 ** This program is distributed in the hope that it will be useful,
 ** but WITHOUT ANY WARRANTY; without even the implied warranty of
 ** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-** GNU Lesser General Public License for more details.
+** GNU General Public License for more details.
 **
 ** You should have received a copy of the GNU General Public License
 ** along with this program; if not, write to the Free Software
@@ -48,6 +48,10 @@
 #define MSC_SERIAL 0x00
 #endif
 
+#ifndef BTN_TOOL_DOUBLETAP
+#define BTN_TOOL_DOUBLETAP 0x14d
+#endif
+
 /*****************************************************************************
 ** Defines
 *****************************************************************************/
@@ -211,17 +215,20 @@ static int USBIdentifyModel(USBTABLET* pUSB);
 
 	static USBSUBTYPE xCintiq[] =
 	{
-		{ "MODEL_PL400",   "Wacom PL400",    1, 0x30 },
-		{ "MODEL_PL500",   "Wacom PL500",    2, 0x31 },
-		{ "MODEL_PL600",   "Wacom PL600",    3, 0x32 },
-		{ "MODEL_PL600SX", "Wacom PL600SX",  4, 0x33 },
-		{ "MODEL_PL550",   "Wacom PL550",    5, 0x34 },
-		{ "MODEL_PL800",   "Wacom PL800",    6, 0x35 },
-		{ "MODEL_PL700",   "Wacom PL700",    7, 0x37 },
-		{ "MODEL_PL510",   "Wacom PL510",    8, 0x38 },
-		{ "MODEL_DTU710",  "Wacom PL710",    9, 0x39 },
-		{ "MODEL_DTF720",  "Wacom DTF720",  10, 0xC0 },
-		{ "MODEL_DTF521",  "Wacom DTF521",  11, 0xC4 },
+		{ "PL400",   "Wacom PL400",    1, 0x30 },
+		{ "PL500",   "Wacom PL500",    2, 0x31 },
+		{ "PL600",   "Wacom PL600",    3, 0x32 },
+		{ "PL600SX", "Wacom PL600SX",  4, 0x33 },
+		{ "PL550",   "Wacom PL550",    5, 0x34 },
+		{ "PL800",   "Wacom PL800",    6, 0x35 },
+		{ "PL700",   "Wacom PL700",    7, 0x37 },
+		{ "PL510",   "Wacom PL510",    8, 0x38 },
+		{ "DTU710",  "Wacom PL710",    9, 0x39 },
+		{ "DTF720",  "Wacom DTF720",  10, 0xC0 },
+		{ "DTF521",  "Wacom DTF521",  11, 0xC4 },
+		{ "DTU1931", "Wacom DTU1931", 12, 0xC7 },
+		{ "DTU2231", "Wacom DTU2231", 13, 0xCE },
+		{ "DTU1631", "Wacom DTU1631", 14, 0xF0 },
 		{ NULL }
 	};
 
@@ -257,9 +264,10 @@ static int USBIdentifyModel(USBTABLET* pUSB);
 
 	static USBSUBTYPE xBamboo[] =
 	{
-		{ "MTE_450", "Wacom Bamboo", 1, 0x65 },
-		{ "CTE_450", "Wacom BambooFun 4x5", 2, 0x17 },
-		{ "CTE_650", "Wacom BambooFun 6x8", 3, 0x18 },
+		{ "MTE_450", "Wacom Bamboo",         1, 0x65 },
+		{ "CTE_450", "Wacom BambooFun 4x5",  2, 0x17 },
+		{ "CTE_650", "Wacom BambooFun 6x8",  3, 0x18 },
+		{ "CTE_631", "Wacom Bamboo1 Medium", 4, 0x19 },
 		{ NULL }
 	};
 
@@ -269,11 +277,20 @@ static int USBIdentifyModel(USBTABLET* pUSB);
 		{ NULL }
 	};
 
+	static USBSUBTYPE xTabletPC[] =
+	{
+		{ "TPC-090", "Wacom Tablet PC90", 1, 0x90 },
+		{ "TPC-093", "Wacom Tablet PC93", 2, 0x93 },
+		{ "TPC-09A", "Wacom Tablet PC9A", 3, 0x9A },
+		{ NULL }
+	};
+
 	static USBSUBTYPE xCintiqV5[] =
 	{
 		{ "DTZ-21ux",  "Wacom Cintiq 21UX",  1, 0x3F },
 		{ "DTZ-20wsx", "Wacom Cintiq 20WSX", 2, 0xC5 },
 		{ "DTZ-12wx",  "Wacom Cintiq 12WX",  3, 0xC6 },
+		{ "DTZ-21ux2", "Wacom Cintiq 21UX2", 4, 0xCC },
 		{ NULL }
 	};
 
@@ -289,6 +306,17 @@ static int USBIdentifyModel(USBTABLET* pUSB);
 		{ NULL }
 	};
 
+	static USBSUBTYPE xIntuos4[] =
+	{
+		{ "PTK-440",   "Wacom Intuos4 4x6",   1, 0xB8 },
+		{ "PTK-640",   "Wacom Intuos4 6x9",   2, 0xB9 },
+		{ "PTK-840",   "Wacom Intuos4 8x13",  3, 0xBA },
+		{ "PTK-1240",  "Wacom Intuos4 12x19", 4, 0xBB },
+		{ "PTK-540WL", "Wacom Intuos4 WLUSB", 5, 0xBC },
+		{ "PTK-540WL", "Wacom Intuos4 WLBT",  6, 0xBD },
+		{ NULL }
+	};
+
 	static USBDEVICE xWacomDevices[] =
 	{
 		{ "pp", "PenPartner", WACOMDEVICE_PENPARTNER, xPenPartner, 1 },
@@ -299,12 +327,14 @@ static int USBIdentifyModel(USBTABLET* pUSB);
 		{ "int", "Intuos", WACOMDEVICE_INTUOS, xIntuos, 2 },
 		{ "int2", "Intuos2", WACOMDEVICE_INTUOS2, xIntuos2, 2 },
 		{ "int3", "Intuos3", WACOMDEVICE_INTUOS3, xIntuos3, 2 },
+		{ "int4", "Intuos4", WACOMDEVICE_INTUOS3, xIntuos4, 2 },
 		{ "ctq", "Cintiq (V5)", WACOMDEVICE_CINTIQV5, xCintiqV5, 2 },
 		{ "pl", "Cintiq (PL)", WACOMDEVICE_CINTIQ, xCintiq, 1 },
 		{ "ptu", "Cintiq Partner (PTU)", WACOMDEVICE_PTU, xCintiqPartner, 1 },
 		{ "vol", "Volito", WACOMDEVICE_VOLITO, xVolito, 1 },
 		{ "vol2", "Volito2", WACOMDEVICE_VOLITO2, xVolito2, 1 },
 		{ "mo", "Bamboo", WACOMDEVICE_MO, xBamboo, 2 },
+		{ "tpc", "Tablet PC", WACOMDEVICE_TPC, xTabletPC, 1 },
 		{ NULL }
 	};
 
@@ -627,7 +657,8 @@ static int USBIdentifyModel(USBTABLET* pUSB)
 				ISBITSET(keybits,BTN_TOOL_AIRBRUSH) ||
 				ISBITSET(keybits,BTN_TOOL_FINGER) ||
 				ISBITSET(keybits,BTN_TOOL_MOUSE) ||
-				ISBITSET(keybits,BTN_TOOL_LENS))
+				ISBITSET(keybits,BTN_TOOL_LENS) ||
+				ISBITSET(keybits,BTN_TOOL_DOUBLETAP))
 			pUSB->state[0].uValid |= BIT(WACOMFIELD_PROXIMITY) |
 					BIT(WACOMFIELD_TOOLTYPE);
 
@@ -777,6 +808,7 @@ static int USBParseKEY(USBTABLET* pUSB, struct input_event* pEv)
 		case BTN_TOOL_MOUSE: tool = WACOMTOOLTYPE_MOUSE; break;
 		case BTN_TOOL_FINGER: tool = WACOMTOOLTYPE_PAD; break;
 		case BTN_TOOL_LENS: tool = WACOMTOOLTYPE_LENS; break;
+		case BTN_TOOL_DOUBLETAP: tool = WACOMTOOLTYPE_TOUCH; break;
 		default:
 			for (i = 0; i < gNumPadKeys; i++)
 				if (pEv->code == gPadKeys [i])
