@@ -1,6 +1,6 @@
 /*
  * Copyright 1995-2002 by Frederic Lepied, France. <Lepied@XFree86.org>
- * Copyright 2002-2008 by Ping Cheng, Wacom Technology. <pingc@wacom.com>		
+ * Copyright 2002-2010 by Ping Cheng, Wacom. <pingc@wacom.com>		
  * 
  * This program is free software; you can redistribute it and/or
  * modify it under the terms of the GNU General Public License
@@ -10,7 +10,7 @@
  * This program is distributed in the hope that it will be useful,
  * but WITHOUT ANY WARRANTY; without even the implied warranty of
  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
- * GNU Lesser General Public License for more details.
+ * GNU General Public License for more details.
  *
  * You should have received a copy of the GNU General Public License
  * along with this program; if not, write to the Free Software 
@@ -25,7 +25,7 @@
 static Bool serialDetect(LocalDevicePtr pDev);
 static Bool serialInit(LocalDevicePtr pDev, char* id, float *version);
 
-static int serialInitTablet(LocalDevicePtr local, char* id, float *version);
+static int serialInitTablet(LocalDevicePtr pDev, char* id, float *version);
 static void serialInitIntuos(WacomCommonPtr common, const char* id, float version);
 static void serialInitIntuos2(WacomCommonPtr common, const char* id, float version);
 static void serialInitCintiq(WacomCommonPtr common, const char* id, float version);
@@ -255,7 +255,7 @@ int xf86WcmWriteWait(int fd, const char* request)
 		len = xf86WriteSerial(fd, request, strlen(request));
 		if ((len == -1) && (errno != EAGAIN))
 		{
-			ErrorF("Wacom xf86WcmWrite error : %s", strerror(errno));
+			ErrorF("Wacom xf86WcmWriteWait error : %s", strerror(errno));
 			return 0;
 		}
 
@@ -700,7 +700,7 @@ static int serialParseProtocol5(LocalDevicePtr local, const unsigned char* data)
 	/* Protocol 5 devices support 2 data channels */
 	channel = data[0] & 0x01;
 
-	/* pick up where we left off, minus relative values */
+	 /* pick up where we left off, minus relative values */
 	ds = &common->wcmChannel[channel].work;
 	RESET_RELATIVE(*ds);
 
@@ -950,6 +950,9 @@ static void serialInitCintiq(WacomCommonPtr common, const char* id, float versio
 		/* PL-800  */
 		common->wcmMaxZ = 511;
 	}
+
+	/* tilt disabled */
+	common->wcmFlags &= ~TILT_ENABLED_FLAG;
 }
 
 static void serialInitPenPartner(WacomCommonPtr common, const char* id, float version)
@@ -961,6 +964,9 @@ static void serialInitPenPartner(WacomCommonPtr common, const char* id, float ve
 	common->wcmMaxZ = 255;
 	common->wcmResolX = 1000; /* tablet X resolution in points/inch */
 	common->wcmResolY = 1000; /* tablet Y resolution in points/inch */
+
+	/* tilt disabled */
+	common->wcmFlags &= ~TILT_ENABLED_FLAG;
 }
 
 static void serialInitGraphire(WacomCommonPtr common, const char* id, float version)
@@ -975,6 +981,9 @@ static void serialInitGraphire(WacomCommonPtr common, const char* id, float vers
 	common->wcmMaxZ = 511;
 	common->wcmResolX = 1016; /* tablet X resolution in points/inch */
 	common->wcmResolY = 1016; /* tablet Y resolution in points/inch */
+
+	/* tilt disabled */
+	common->wcmFlags &= ~TILT_ENABLED_FLAG;
 }
 
 static void serialInitProtocol4(WacomCommonPtr common, const char* id, float version)
@@ -992,6 +1001,9 @@ static void serialInitProtocol4(WacomCommonPtr common, const char* id, float ver
 		else
 			common->wcmMaxZ = 120;
 	}
+
+	/* tilt disabled */
+	common->wcmFlags &= ~TILT_ENABLED_FLAG;
 }
 
 static void serialGetResolution(LocalDevicePtr local)
@@ -1249,16 +1261,15 @@ static void serialParseP4Common(LocalDevicePtr local,
 	if (!last->proximity && ds->proximity)
 		ds->device_type = cur_type;
 	/* check on previous proximity */
-	else if (is_stylus && ds->proximity)
+	else if (ds->buttons && ds->proximity)
 	{
-		/* we were fooled by tip and second
+		/* we might have been fooled by tip and second
 		 * sideswitch when it came into prox */
 		if ((ds->device_type != cur_type) &&
 			(ds->device_type == ERASER_ID))
 		{
 			/* send a prox-out for old device */
-			WacomDeviceState out = { 0 };
-			xf86WcmEvent(common, 0, &out);
+			xf86WcmSoftOut(common, 0);
 			ds->device_type = cur_type;
 		}
 	}
@@ -1295,9 +1306,11 @@ int xf86WcmSerialValidate(WacomCommonPtr common, const unsigned char* data)
 				((i!=0) && (data[i] & HEADER_BIT)) )
 		{
 			bad = 1;
-			ErrorF("xf86WcmSerialValidate: bad magic at %d "
-				"v=%x l=%d\n", i, data[i], common->wcmPktLength);
-			if (i!=0 && (data[i] & HEADER_BIT)) return i;
+			if (i!=0 && (data[i] & HEADER_BIT)) {
+				xf86Msg(X_WARNING, "xf86WcmSerialValidate: bad magic at %d "
+					"v=%x l=%d\n", i, data[i], common->wcmPktLength);
+				return i;
+			}
 		}
 	}
 	if (bad) return common->wcmPktLength;
