@@ -1,6 +1,6 @@
 /*
  * Copyright 1995-2002 by Frederic Lepied, France. <Lepied@XFree86.org>
- * Copyright 2002-2008 by Ping Cheng, Wacom Technology. <pingc@wacom.com>		
+ * Copyright 2002-2010 by Ping Cheng, Wacom Technology. <pingc@wacom.com>		
  *                                                                            
  * This program is free software; you can redistribute it and/or
  * modify it under the terms of the GNU General Public License
@@ -10,7 +10,7 @@
  * This program is distributed in the hope that it will be useful,
  * but WITHOUT ANY WARRANTY; without even the implied warranty of
  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
- * GNU Lesser General Public License for more details.
+ * GNU General Public License for more details.
  *
  * You should have received a copy of the GNU General Public License
  * along with this program; if not, write to the Free Software 
@@ -24,27 +24,26 @@
 
 #include <sys/utsname.h>
 
-/* support for compiling module on kernels older than 2.6 */
-#ifndef EV_MSC
-#define EV_MSC 0x04
-#endif
-
-#ifndef MSC_SERIAL
-#define MSC_SERIAL 0x00
-#endif
-
-#ifndef EV_SYN
-#define EV_SYN 0x00
-#endif
-
-#ifndef SYN_REPORT
-#define SYN_REPORT 0
-#endif
-
+/* Defines on newer kernels */
 #ifndef BTN_TASK
 #define BTN_TASK 0x117
 #endif
 
+#ifndef BTN_TOOL_TRIPLETAP
+#define BTN_TOOL_TRIPLETAP 0x14e
+#endif
+
+#ifdef WCM_CUSTOM_DEBUG
+	static int lastPenSerial = 0;
+	extern void detectPressureIssue(struct input_event* event,
+		    WacomCommonPtr common, int channel);
+	extern void detectChannelChange(LocalDevicePtr local, int channel);
+	extern void dumpEventRing(LocalDevicePtr local);
+	extern void dumpChannels(LocalDevicePtr local);
+	extern char* timestr();
+	extern void logEvent(const struct input_event* event);
+#endif
+
 static Bool usbDetect(LocalDevicePtr);
 Bool usbWcmInit(LocalDevicePtr pDev, char* id, float *version);
 
@@ -57,7 +56,7 @@ static int usbParse(LocalDevicePtr local, const unsigned char* data);
 static int usbDetectConfig(LocalDevicePtr local);
 static void usbParseEvent(LocalDevicePtr local,
 	const struct input_event* event);
-static void usbParseChannel(LocalDevicePtr local, int channel, int serial);
+static void usbParseChannel(LocalDevicePtr local, int channel);
 
 	WacomDeviceClass gWacomUSBDevice =
 	{
@@ -212,7 +211,7 @@ static void usbParseChannel(LocalDevicePtr local, int channel, int serial);
 
 	static WacomModel usbCintiq =
 	{
-		"USB Cintiq",
+		"USB PL/Cintiq",
 		usbInitProtocol4,
 		NULL,                 /* resolution not queried */
 		usbWcmGetRanges,
@@ -222,7 +221,7 @@ static void usbParseChannel(LocalDevicePtr local, int channel, int serial);
 		NULL,                 /* link speed unsupported */
 		NULL,                 /* start not supported */
 		usbParse,
-		NULL,                 /* input filtering */
+		xf86WcmFilterCoord,   /* input filtering */
 		usbDetectConfig,      /* detect hardware buttons etc */
 	};
 
@@ -238,13 +237,13 @@ static void usbParseChannel(LocalDevicePtr local, int channel, int serial);
 		NULL,                 /* link speed unsupported */
 		NULL,                 /* start not supported */
 		usbParse,
-		NULL,                 /* input filtering */
+		xf86WcmFilterCoord,   /* input filtering */
 		usbDetectConfig,      /* detect hardware buttons etc */
 	};
 
-	static WacomModel usbIntuos =
+	static WacomModel usbIntuos1 =
 	{
-		"USB Intuos",
+		"USB Intuos1",
 		usbInitProtocol5,
 		NULL,                 /* resolution not queried */
 		usbWcmGetRanges,
@@ -290,6 +289,22 @@ static void usbParseChannel(LocalDevicePtr local, int channel, int serial);
 		usbDetectConfig,      /* detect hardware buttons etc */
 	};
 
+	static WacomModel usbIntuos4 =
+	{
+		"USB Intuos4",
+		usbInitProtocol5,
+		NULL,                 /* resolution not queried */
+		usbWcmGetRanges,
+		NULL,                 /* reset not supported */
+		NULL,                 /* tilt automatically enabled */
+		NULL,                 /* suppress implemented in software */
+		NULL,                 /* link speed unsupported */
+		NULL,                 /* start not supported */
+		usbParse,
+		xf86WcmFilterIntuos,  /* input filtering recommended */
+		usbDetectConfig,      /* detect hardware buttons etc */
+	};
+
 	static WacomModel usbVolito =
 	{
 		"USB Volito",
@@ -338,6 +353,22 @@ static void usbParseChannel(LocalDevicePtr local, int channel, int serial);
 		usbDetectConfig,      /* detect hardware buttons etc */
 	};
 
+	static WacomModel usbTabletPC =
+	{
+		"USB TabletPC",
+		usbInitProtocol4,
+		NULL,                 /* resolution not queried */
+		usbWcmGetRanges,
+		NULL,                 /* reset not supported */
+		NULL,                 /* tilt automatically enabled */
+		NULL,                 /* suppress implemented in software */
+		NULL,                 /* link speed unsupported */
+		NULL,                 /* start not supported */
+		usbParse,
+		xf86WcmFilterCoord,   /* input filtering */
+		usbDetectConfig,      /* detect hardware buttons etc */
+	};
+
 /*****************************************************************************
  * usbDetect --
  *   Test if the attached device is USB.
@@ -355,7 +386,7 @@ static Bool usbDetect(LocalDevicePtr local)
 
 	if (err < 0)
 	{
-		ErrorF("usbDetect: can not ioctl version\n");
+		xf86Msg(X_WARNING, "usbDetect: can not ioctl version\n");
 		return 0;
 	}
 #ifdef EVIOCGRAB
@@ -365,8 +396,6 @@ static Bool usbDetect(LocalDevicePtr local)
 	if (err < 0) 
 		ErrorF("%s Wacom X driver can't grab event device, errno=%d\n",
 				local->name, errno);
-	else 
-		ErrorF("%s Wacom X driver grabbed event device\n", local->name);
 #endif
 	return 1;
 }
@@ -374,12 +403,6 @@ static Bool usbDetect(LocalDevicePtr local)
 /*****************************************************************************
  * wcmusbInit --
  ****************************************************************************/
-#define BIT(x)		(1<<((x) & (BITS_PER_LONG - 1)))
-#define BITS_PER_LONG	(sizeof(long) * 8)
-#define NBITS(x)	((((x)-1)/BITS_PER_LONG)+1)
-#define ISBITSET(x,y)	((x)[LONG(y)] & BIT(y))
-#define OFF(x)		((x)%BITS_PER_LONG)
-#define LONG(x)		((x)/BITS_PER_LONG)
 
 /* Key codes used to mark tablet buttons -- must be in sync
  * with the keycode array in wacom.c kernel driver.
@@ -411,13 +434,14 @@ static struct
 	{ 0x16, 2032, 2032, &usbGraphire4  }, /* Graphire4 6x8 */ 
 	{ 0x17, 2540, 2540, &usbBambooFun  }, /* BambooFun 4x5 */
 	{ 0x18, 2540, 2540, &usbBambooFun  }, /* BambooFun 6x8 */
+	{ 0x19, 2032, 2032, &usbBamboo1    }, /* Bamboo1 Medium*/ 
 	{ 0x81, 2032, 2032, &usbGraphire4  }, /* Graphire4 6x8 BlueTooth */
 
-	{ 0x20, 2540, 2540, &usbIntuos     }, /* Intuos 4x5 */
-	{ 0x21, 2540, 2540, &usbIntuos     }, /* Intuos 6x8 */
-	{ 0x22, 2540, 2540, &usbIntuos     }, /* Intuos 9x12 */
-	{ 0x23, 2540, 2540, &usbIntuos     }, /* Intuos 12x12 */
-	{ 0x24, 2540, 2540, &usbIntuos     }, /* Intuos 12x18 */
+	{ 0x20, 2540, 2540, &usbIntuos1    }, /* Intuos 4x5 */
+	{ 0x21, 2540, 2540, &usbIntuos1    }, /* Intuos 6x8 */
+	{ 0x22, 2540, 2540, &usbIntuos1    }, /* Intuos 9x12 */
+	{ 0x23, 2540, 2540, &usbIntuos1    }, /* Intuos 12x12 */
+	{ 0x24, 2540, 2540, &usbIntuos1    }, /* Intuos 12x18 */
 
 	{ 0x03,  508,  508, &usbCintiqPartner }, /* PTU600 */
 
@@ -431,7 +455,11 @@ static struct
 	{ 0x38,  508,  508, &usbCintiq     }, /* PL510 */
 	{ 0x39,  508,  508, &usbCintiq     }, /* PL710 */ 
 	{ 0xC0,  508,  508, &usbCintiq     }, /* DTF720 */
+	{ 0xC2,  508,  508, &usbCintiq     }, /* DTF720a */
 	{ 0xC4,  508,  508, &usbCintiq     }, /* DTF521 */ 
+	{ 0xC7, 2540, 2540, &usbCintiq     }, /* DTU1931 */
+	{ 0xCE, 2540, 2540, &usbCintiq     }, /* DTU2231 */
+	{ 0xF0, 2540, 2540, &usbCintiq     }, /* DTU1631 */
 
 	{ 0x41, 2540, 2540, &usbIntuos2    }, /* Intuos2 4x5 */
 	{ 0x42, 2540, 2540, &usbIntuos2    }, /* Intuos2 6x8 */
@@ -449,6 +477,11 @@ static struct
 
 	{ 0x65, 2540, 2540, &usbBamboo     }, /* Bamboo */
 	{ 0x69, 1012, 1012, &usbBamboo1    }, /* Bamboo1 */ 
+	{ 0xD1, 2540, 2540, &usbBamboo     }, /* CTL-460 */
+	{ 0xD4, 2540, 2540, &usbBamboo     }, /* CTH-461 */
+	{ 0xD3, 2540, 2540, &usbBamboo     }, /* CTL-660 */
+	{ 0xD2, 2540, 2540, &usbBamboo     }, /* CTL-461/S */
+	{ 0xD0, 2540, 2540, &usbBamboo     }, /* Bamboo Touch */
 
 	{ 0xB0, 5080, 5080, &usbIntuos3    }, /* Intuos3 4x5 */
 	{ 0xB1, 5080, 5080, &usbIntuos3    }, /* Intuos3 6x8 */
@@ -458,38 +491,85 @@ static struct
 	{ 0xB5, 5080, 5080, &usbIntuos3    }, /* Intuos3 6x11 */
 	{ 0xB7, 5080, 5080, &usbIntuos3    }, /* Intuos3 4x6 */
 
+	{ 0xB8, 5080, 5080, &usbIntuos4    }, /* Intuos4 4x6 */
+	{ 0xB9, 5080, 5080, &usbIntuos4    }, /* Intuos4 6x9 */
+	{ 0xBA, 5080, 5080, &usbIntuos4    }, /* Intuos4 8x13 */
+	{ 0xBB, 5080, 5080, &usbIntuos4    }, /* Intuos4 12x19*/
+	{ 0xBC, 5080, 5080, &usbIntuos4    }, /* Intuos4 WL USB Endpoint */
+	{ 0xBD, 5080, 5080, &usbIntuos4    }, /* Intuos4 WL Bluetooth Endpoint */
+
 	{ 0x3F, 5080, 5080, &usbCintiqV5   }, /* Cintiq 21UX */ 
 	{ 0xC5, 5080, 5080, &usbCintiqV5   }, /* Cintiq 20WSX */ 
-	{ 0xC6, 5080, 5080, &usbCintiqV5   }  /* Cintiq 12WX */ 
+	{ 0xC6, 5080, 5080, &usbCintiqV5   }, /* Cintiq 12WX */ 
+	{ 0xCC, 5080, 5080, &usbCintiqV5   }, /* Cintiq 21UX2 */ 
+
+	{ 0x90, 2540, 2540, &usbTabletPC   }, /* TabletPC 0x90 */ 
+	{ 0x93, 2540, 2540, &usbTabletPC   }, /* TabletPC 0x93 */
+	{ 0x9A, 2540, 2540, &usbTabletPC   }, /* TabletPC 0x9A */
+	{ 0x9F,   10,   10, &usbTabletPC   }, /* CapPlus  0x9F */
+	{ 0xE2,   10,   10, &usbTabletPC   }, /* TabletPC 0xE2 */ 
+	{ 0xE3, 2540, 2540, &usbTabletPC   }  /* TabletPC 0xE3 */
 };
 
+static void usbRetrieveKeys(WacomCommonPtr common)
+{
+	int i;
+
+	ioctl(common->fd, EVIOCGBIT(EV_KEY, sizeof(common->wcmKeys)), common->wcmKeys);
+
+	/* Find out supported button codes - except mouse button codes
+	 * BTN_LEFT and BTN_RIGHT, which are always fixed. */
+	common->npadkeys = 0;
+	for (i = 0; i < sizeof (padkey_codes) / sizeof (padkey_codes [0]); i++)
+		if (ISBITSET (common->wcmKeys, padkey_codes [i]))
+			common->padkey_code [common->npadkeys++] = padkey_codes [i];
+
+	/* set default nbuttons */
+	if (ISBITSET (common->wcmKeys, BTN_TASK))
+		common->nbuttons = 10;
+	else if (ISBITSET (common->wcmKeys, BTN_BACK))
+		common->nbuttons = 9;
+	else if (ISBITSET (common->wcmKeys, BTN_FORWARD))
+		common->nbuttons = 8;
+	else if (ISBITSET (common->wcmKeys, BTN_EXTRA))
+		common->nbuttons = 7;
+	else if (ISBITSET (common->wcmKeys, BTN_SIDE))
+		common->nbuttons = 6;
+	else
+		common->nbuttons = 5;
+}
+
 Bool usbWcmInit(LocalDevicePtr local, char* id, float *version)
 {
 	int i;
-	short sID[4];
-	unsigned long keys[NBITS(KEY_MAX)];
 	WacomDevicePtr priv = (WacomDevicePtr)local->private;
 	WacomCommonPtr common = priv->common;
 
 	DBG(1, priv->debugLevel, ErrorF("initializing USB tablet\n"));
 	*version = 0.0;
 
-	/* fetch vendor, product, and model name */
-	ioctl(local->fd, EVIOCGID, sID);
+	/* fetch model name */
 	ioctl(local->fd, EVIOCGNAME(sizeof(id)), id);
 
-	/* vendor is wacom */
-	if (sID[1] == 0x056A)
-	{
-		common->tablet_id = sID[2];
+	/* search for the proper device id */
+	for (i = 0; i < sizeof (WacomModelDesc) / sizeof (WacomModelDesc [0]); i++)
+		if (common->tablet_id == WacomModelDesc [i].model_id)
+		{
+			common->wcmModel = WacomModelDesc [i].model;
+			common->wcmResolX = WacomModelDesc [i].xRes;
+			common->wcmResolY = WacomModelDesc [i].yRes;
+		}
 
-		for (i = 0; i < sizeof (WacomModelDesc) / sizeof (WacomModelDesc [0]); i++)
-			if (common->tablet_id == WacomModelDesc [i].model_id)
-			{
-				common->wcmModel = WacomModelDesc [i].model;
-				common->wcmResolX = WacomModelDesc [i].xRes;
-				common->wcmResolY = WacomModelDesc [i].yRes;
-			}
+	if (common->wcmModel && strstr(common->wcmModel->name, "TabletPC"))
+	{
+		/* Tablet PC button applied to the whole tablet. Not just one tool */
+		common->wcmTPCButtonDefault = 1;
+	}
+
+	if ( priv->flags & STYLUS_ID )
+	{
+		common->wcmTPCButton = xf86SetBoolOption(local->options, 
+			"TPCButton", common->wcmTPCButtonDefault);
 	}
 
 	if (!common->wcmModel)
@@ -498,35 +578,16 @@ Bool usbWcmInit(LocalDevicePtr local, char* id, float *version)
 		common->wcmResolX = common->wcmResolY = 1016;
 	}
 
-	/* Determine max number of buttons */
-	if (ioctl(local->fd, EVIOCGBIT(EV_KEY,sizeof(keys)),keys) < 0)
-	{
-		ErrorF("WACOM: unable to ioctl key bits.\n");
-		return FALSE;
-	}
-
-	/* Find out supported button codes - except mouse button codes
-	 * BTN_LEFT and BTN_RIGHT, which are always fixed. */
-	common->npadkeys = 0;
-	for (i = 0; i < sizeof (padkey_codes) / sizeof (padkey_codes [0]); i++)
-		if (ISBITSET (keys, padkey_codes [i]))
-			common->padkey_code [common->npadkeys++] = padkey_codes [i];
-
-	if (ISBITSET (keys, BTN_TASK))
-		common->nbuttons = 10;
-	else if (ISBITSET (keys, BTN_BACK))
-		common->nbuttons = 9;
-	else if (ISBITSET (keys, BTN_FORWARD))
-		common->nbuttons = 8;
-	else if (ISBITSET (keys, BTN_EXTRA))
-		common->nbuttons = 7;
-	else if (ISBITSET (keys, BTN_SIDE))
-		common->nbuttons = 6;
-	else
-		common->nbuttons = 5;
-
-	common->wcmFlags |= TILT_ENABLED_FLAG;
+	/* we have to call this ioclt again since on some older kernels the first time
+	 * when system reboot, we do not get all keys. Looks like kernel 2.6.27 and 
+	 * later work all right.
+	 */
+	usbRetrieveKeys(common);
 
+	/* Use the default nbuttons and npadkeys since the EVIOCGBIT does not always
+	 * return the correct number of keys/buttons
+	*/
+	common->nbuttons =  common->npadkeys = MAX_BUTTONS;
 	return Success;
 }
 
@@ -537,6 +598,10 @@ static void usbInitProtocol5(WacomCommonPtr common, const char* id,
 	common->wcmPktLength = sizeof(struct input_event);
 	common->wcmCursorProxoutDistDefault 
 			= PROXOUT_INTUOS_DISTANCE;
+
+	/* tilt enabled */
+	common->wcmFlags |= TILT_ENABLED_FLAG;
+
 	/* reinitialize max here since 0 is for Graphire series */
 	common->wcmMaxCursorDist = 256;
 
@@ -549,104 +614,112 @@ static void usbInitProtocol4(WacomCommonPtr common, const char* id,
 	common->wcmPktLength = sizeof(struct input_event);
 	common->wcmCursorProxoutDistDefault 
 			= PROXOUT_GRAPHIRE_DISTANCE;
+
+	/* tilt disabled */
+	common->wcmFlags &= ~TILT_ENABLED_FLAG;
 }
 
 int usbWcmGetRanges(LocalDevicePtr local)
 {
-	int nValues[5];
+	struct input_absinfo absinfo;
 	unsigned long ev[NBITS(EV_MAX)];
 	unsigned long abs[NBITS(ABS_MAX)];
 	WacomDevicePtr priv = (WacomDevicePtr)local->private;
 	WacomCommonPtr common =	priv->common;
+	int is_touch;
 
-	if (ioctl(local->fd, EVIOCGBIT(0 /*EV*/, sizeof(ev)), ev) < 0)
+	is_touch = IsTouch(priv);
+
+	/* Devices, such as Bamboo P&T, may have Pad data reported in the same
+	 * packet as Touch.  Its normal for Pad to be called first but logic
+	 * requires it to act the same as Touch.
+	 */
+	if (ISBITSET(common->wcmKeys, BTN_TOOL_DOUBLETAP) &&
+	    ISBITSET(common->wcmKeys, BTN_TOOL_FINGER) && 
+	    (common->tablet_id >= 0xd0 && common->tablet_id <= 0xd3))
+		is_touch = 1;
+
+	if (ioctl(local->fd, EVIOCGBIT(EV_SYN, sizeof(ev)), ev) < 0)
 	{
 		ErrorF("WACOM: unable to ioctl event bits.\n");
 		return !Success;
 	}
 
-	/* determine if this version of the kernel uses SYN_REPORT or MSC_SERIAL
-	   to indicate end of record for a complete tablet event */
-	if (ISBITSET(ev,EV_SYN))
-	{
-		common->wcmFlags |= USE_SYN_REPORTS_FLAG;
-	}
-	else
-	{
-		ErrorF("WACOM: Kernel doesn't support SYN_REPORT\n");
-		common->wcmFlags &= ~USE_SYN_REPORTS_FLAG;
-	}
+	common->wcmFlags |= USE_SYN_REPORTS_FLAG;
 
-	if (ioctl(local->fd, EVIOCGBIT(EV_ABS,sizeof(abs)),abs) < 0)
+	if (ioctl(local->fd, EVIOCGBIT(EV_ABS, sizeof(abs)), abs) < 0)
 	{
 		ErrorF("WACOM: unable to ioctl abs bits.\n");
 		return !Success;
 	}
 
 	/* absolute values */
-	if (!ISBITSET(ev,EV_ABS))
+	if (!ISBITSET(ev, EV_ABS))
 	{
 		ErrorF("WACOM: unable to ioctl max values.\n");
 		return !Success;
 	}
 
 	/* max x */
-	if (ioctl(local->fd, EVIOCGABS(ABS_X), nValues) < 0)
+	if (ioctl(local->fd, EVIOCGABS(ABS_X), &absinfo) < 0)
 	{
 		ErrorF("WACOM: unable to ioctl xmax value.\n");
 		return !Success;
 	}
-	common->wcmMaxX = nValues[2];
-	if (common->wcmMaxX <= 0)
-	{
-		ErrorF("WACOM: xmax value is wrong.\n");
-		return !Success;
-	}
+	if (!is_touch)
+		common->wcmMaxX = absinfo.maximum;
+	else
+		common->wcmMaxTouchX = absinfo.maximum;
 
 	/* max y */
-	if (ioctl(local->fd, EVIOCGABS(ABS_Y), nValues) < 0)
+	if (ioctl(local->fd, EVIOCGABS(ABS_Y), &absinfo) < 0)
 	{
 		ErrorF("WACOM: unable to ioctl ymax value.\n");
 		return !Success;
 	}
-	common->wcmMaxY = nValues[2];
-	if (common->wcmMaxY <= 0)
+	if (!is_touch)
+		common->wcmMaxY = absinfo.maximum;
+	else
+		common->wcmMaxTouchY = absinfo.maximum;
+
+	/* max finger strip X for tablets with Expresskeys 
+	 * or touch physical X for TabletPCs with touch */
+	if (ioctl(local->fd, EVIOCGABS(ABS_RX), &absinfo) == 0)
 	{
-		ErrorF("WACOM: ymax value is wrong.\n");
-		return !Success;
+		if (is_touch)
+			common->wcmTouchResolX = absinfo.maximum;
+		else
+			common->wcmMaxStripX = absinfo.maximum;
+
+		DBG(3, common->debugLevel, ErrorF("%s - usbWcmGetRanges: Got ABS_RX %d\n", 
+				local->name, absinfo.maximum));
+	}
+
+	/* max finger strip y for tablets with Expresskeys
+	 * or touch physical Y for TabletPCs with touch */
+	if (ioctl(local->fd, EVIOCGABS(ABS_RY), &absinfo) == 0)
+	{
+		if (is_touch)
+			common->wcmTouchResolY = absinfo.maximum;
+		else
+			common->wcmMaxStripY = absinfo.maximum;
+	}
+
+	if (is_touch && common->wcmMaxTouchX && common->wcmTouchResolX)
+	{
+		common->wcmTouchResolX = (int)(((double)common->wcmTouchResolX)
+			 / ((double)common->wcmMaxTouchX) + 0.5);
+		common->wcmTouchResolY = (int)(((double)common->wcmTouchResolY)
+			 / ((double)common->wcmMaxTouchY) + 0.5);
 	}
 
 	/* max z cannot be configured */
-	if (ioctl(local->fd, EVIOCGABS(ABS_PRESSURE), nValues) < 0)
-	{
-		ErrorF("WACOM: unable to ioctl press max value.\n");
-		return !Success;
-	}
-	common->wcmMaxZ = nValues[2];
-	if (common->wcmMaxZ <= 0)
-	{
-		ErrorF("WACOM: press max value is wrong.\n");
-		return !Success;
-	}
+	if (ioctl(local->fd, EVIOCGABS(ABS_PRESSURE), &absinfo) == 0)
+		common->wcmMaxZ = absinfo.maximum;
 
 	/* max distance */
-	if (ioctl(local->fd, EVIOCGABS(ABS_DISTANCE), nValues) < 0)
-	{
-		ErrorF("WACOM: unable to ioctl press max distance.\n");
-		return !Success;
-	}
-	common->wcmMaxDist = nValues[2];
-	if (common->wcmMaxDist < 0)
-	{
-		ErrorF("WACOM: max distance value is wrong.\n");
-		return !Success;
-	}
-
-	/* max fingerstrip X */
-	if (ioctl(local->fd, EVIOCGABS(ABS_RX), nValues) == 0)
-		common->wcmMaxStripX = nValues[2];
-	if (ioctl(local->fd, EVIOCGABS(ABS_RY), nValues) == 0)
-		common->wcmMaxStripY = nValues[2];
+	if (ioctl(local->fd, EVIOCGABS(ABS_DISTANCE), &absinfo) == 0)
+		common->wcmMaxDist = absinfo.maximum;
 
 	return Success;
 }
@@ -657,27 +730,9 @@ static int usbDetectConfig(LocalDevicePtr local)
 	WacomCommonPtr common = priv->common;
 
 	DBG(10, common->debugLevel, ErrorF("usbDetectConfig \n"));
-	if (IsPad (priv))
-	{
-		priv->nbuttons = common->npadkeys;
 
-/* This code will be used when we are ready to report valuators in tablet and tool 
- * specific form, whcih will need to clean InitValuatorAxisStruct() in xf86Wacom.c
- * and all the calls to X related to valuators, such as xf86PostButtonEvent and 
- * xf86PostButtonEvent, etc. Code under util will need to be updaed as well.
- * This will need some time. We put it in the to-do list for now.  Ping 
-		unsigned long abs[NBITS(ABS_MAX)];
-		priv->naxes = 0;
-		if (ioctl(local->fd, EVIOCGBIT(EV_ABS, sizeof(abs)), abs) >= 0)
-		{
-			if (ISBITSET (abs, ABS_RX))
-				priv->naxes++;
-			if (ISBITSET (abs, ABS_RY))
-				priv->naxes++;
-			if (!priv->naxes)
-				priv->flags |= BUTTONS_ONLY_FLAG;
-		}
-*/	}
+	if (IsPad (priv))
+		priv->nbuttons = common->npadkeys;
 	else
 		priv->nbuttons = common->nbuttons;
 
@@ -696,10 +751,115 @@ static int usbParse(LocalDevicePtr local, const unsigned char* data)
 	return common->wcmPktLength;
 }
 
-static void usbParseEvent(LocalDevicePtr local,
-	const struct input_event* event)
+
+static int usbChooseChannel(LocalDevicePtr local)
 {
-	int i, channel;
+	/* figure out the channel to use based on serial number */
+	int i, channel = -1;
+	WacomDevicePtr priv = (WacomDevicePtr)local->private;
+	WacomCommonPtr common = priv->common;
+	int serial = common->wcmLastToolSerial;
+
+	if (common->wcmProtocolLevel == 4)
+	{
+		/* Protocol 4 doesn't support tool serial numbers.
+		 * However, we pass finger index into serial 
+		 * numbers for tablets with multi-touch capabilities 
+		 * to track individual fingers in proper channels.
+		 * serial number 0xf0 is reserved for the pad and is
+		 * always the last supported channel (i.e. MAX_CHANNELS-1).
+		 */
+		if (serial == 0xf0)
+			channel = MAX_CHANNELS-1;
+		else if (serial)
+			channel = serial-1;
+		else
+			channel = 0;
+	}  /* serial number should never be 0 for V5 devices */
+	else if (serial)
+	{
+		/* dual input is supported */
+		if ( strstr(common->wcmModel->name, "Intuos1") || strstr(common->wcmModel->name, "Intuos2") )
+		{
+			/* find existing channel */
+			for (i=0; i<MAX_CHANNELS; ++i)
+			{
+				if (common->wcmChannel[i].work.proximity && 
+			  		common->wcmChannel[i].work.serial_num == serial)
+				{
+					channel = i;
+					break;
+				}
+			}
+
+			/* find an empty channel */
+			if (channel < 0)
+			{
+#ifdef WCM_CUSTOM_DEBUG
+			        dumpChannels(local);
+#endif
+				for (i=0; i<MAX_CHANNELS; ++i)
+				{
+					if (!common->wcmChannel[i].work.proximity)
+					{
+						channel = i;
+						break;
+					}
+				}
+			}
+		}
+		else  /* one transducer plus expresskey (pad) is supported */
+		{
+			if (serial == -1)  /* pad */
+				channel = 1;
+			else if ( (common->wcmChannel[0].work.proximity &&  /* existing transducer */
+				    (common->wcmChannel[0].work.serial_num == serial)) ||
+					!common->wcmChannel[0].work.proximity ) /* new transducer */
+				channel = 0;
+		}
+	}
+
+	/* fresh out of channels */
+	if (channel < 0)
+	{
+		/* This should never happen in normal use.
+		 * Let's start over again. Force prox-out for all channels.
+		 */
+		for (i=0; i<MAX_CHANNELS; ++i)
+		{
+			if (common->wcmChannel[i].work.proximity && 
+					(common->wcmChannel[i].work.serial_num != -1))
+			{
+				common->wcmChannel[i].work.proximity = 0;
+				/* dispatch event */
+				xf86WcmEvent(common, i, &common->wcmChannel[i].work);
+#ifdef WCM_CUSTOM_DEBUG
+				DBG(2, common->debugLevel, ErrorF("%s - usbParse: dropping %d\n", 
+					timestr(), common->wcmChannel[i].work.serial_num));
+				/* we take this channel as a valid one */
+				channel = i;
+#endif
+			}
+		}
+#ifdef WCM_CUSTOM_DEBUG
+		DBG(1, common->debugLevel, ErrorF("%s - usbParse (%s with serial number: %u):"
+			" Exceeded channel count; ignoring the events.\n", 
+			timestr(), local->name, serial));
+		DBG(2, common->debugLevel, dumpEventRing(local));
+#else
+		DBG(1, common->debugLevel, ErrorF("usbParse (device with serial number: %u)"
+			" at %d: Exceeded channel count; ignoring the events.\n", 
+			serial, (int)GetTimeInMillis()));
+#endif
+	}
+
+	return channel;
+}
+
+static void usbParseEvent(LocalDevicePtr local, 
+		const struct input_event* event)
+{
+	int channel = -1;
 	WacomDevicePtr priv = (WacomDevicePtr)local->private;
 	WacomCommonPtr common = priv->common;
 
@@ -708,25 +868,29 @@ static void usbParseEvent(LocalDevicePtr local,
 	 * the serial number; without it we cannot determine the
 	 * correct channel. */
 
+#ifdef WCM_CUSTOM_DEBUG
+	/* save event in ring buffer for debug output later */
+	logEvent(event);
+#endif
 	/* space left? bail if not. */
 	if (common->wcmEventCnt >=
 		(sizeof(common->wcmEvents)/sizeof(*common->wcmEvents)))
 	{
-		ErrorF("usbParse: Exceeded event queue (%d)\n",
+#ifdef WCM_CUSTOM_DEBUG
+		ErrorF("%s - usbParse: Exceeded event queue (%d) \n",
+				timestr(), common->wcmEventCnt);
+#else
+		ErrorF("usbParse: Exceeded event queue (%d) \n",
 				common->wcmEventCnt);
-		common->wcmEventCnt = 0;
-		common->wcmLastToolSerial = 0;
-		return;
+#endif
+		goto skipEvent;
 	}
 
 	/* save it for later */
 	common->wcmEvents[common->wcmEventCnt++] = *event;
 
-	/* Check for end of record indicator.  On 2.4 kernels MSC_SERIAL is used
-	   but on 2.6 and later kernels SYN_REPORT is used.  */
 	if ((event->type == EV_MSC) && (event->code == MSC_SERIAL))
 	{
-		/* save the serial number so we can look up the channel number later */
 		common->wcmLastToolSerial = event->value;
 
 		/* if SYN_REPORT is end of record indicator, we are done */
@@ -734,8 +898,8 @@ static void usbParseEvent(LocalDevicePtr local,
 			return;
 
 		/* fall through to deliver the X event */
-	}
-	else if ((event->type == EV_SYN) && (event->code == SYN_REPORT))
+
+	} else if ((event->type == EV_SYN) && (event->code == SYN_REPORT))
 	{
 		/* if we got a SYN_REPORT but weren't expecting one, change over to
 		   using SYN_REPORT as the end of record indicator */
@@ -747,95 +911,107 @@ static void usbParseEvent(LocalDevicePtr local,
 			common->wcmFlags |= USE_SYN_REPORTS_FLAG;
 		}
 
-		/* fall through to deliver the X event */
+		/* end of record. fall through to deliver the X event */
 	}
 	else
 	{
-		/* not an MSC_SERIAL or SYN_REPORT, bail out */
+		/* not an MSC_SERIAL and not an SYN_REPORT, bail out */
 		return;
 	}
 
-	/* figure out the channel to use based on serial number */
-	channel = -1;
-	if (common->wcmProtocolLevel == 4)
+	/* ignore events without information.
+	 * this case normally happens when a tool other than pad is in
+	 * prox while the expresskey (represented as pad) is pressed
+	 * with the same value (button/ring postion/strip)
+	 */
+	if (common->wcmEventCnt <= 2 && common->wcmLastToolSerial) 
 	{
-		/* Protocol 4 don't support tool serial numbers */
-		if (common->wcmLastToolSerial == 0xf0)
-			channel = 1;
-		else
-			channel = 0;
-		if (!common->wcmChannel[channel].work.proximity)
-		{
-			memset(&common->wcmChannel[channel],0,sizeof(WacomChannel));
-			/* in case the in-prox event was missing */
-			common->wcmChannel[channel].work.proximity = 1;
-		}
-	}
-	else
-	{
-		/* find existing channel */
-		for (i=0; i<MAX_CHANNELS; ++i)
-		{
-			if (common->wcmChannel[i].work.proximity &&
-			   common->wcmChannel[i].work.serial_num == common->wcmLastToolSerial)
-			{
-				channel = i;
-				break;
-			}
-		}
-
-		/* find an empty channel */
-		if (channel < 0)
-		{
-			for (i=0; i<MAX_CHANNELS; ++i)
-			{
-				if (!common->wcmChannel[i].work.proximity)
-				{
-					memset(&common->wcmChannel[i],0,
-							sizeof(WacomChannel));
-					/* in case the in-prox event was missing */
-					common->wcmChannel[i].work.proximity = 1;
-					channel = i;
-					break;
-				}
-			}
-		}
+#ifdef WCM_CUSTOM_DEBUG
+		DBG(3, common->debugLevel, ErrorF("%s - usbParse: dropping empty %s event for serial %d\n", 
+			timestr(), local->name, common->wcmLastToolSerial));
+#else
+		DBG(3, common->debugLevel, ErrorF("%s - usbParse: dropping empty event for serial %d\n", 
+			local->name, common->wcmLastToolSerial));
+#endif
+		goto skipEvent;
 	}
 
-	/* fresh out of channels */
-	if (channel < 0)
-	{
-		/* this should never happen in normal use */
-		ErrorF("usbParse: Exceeded channel count; "
-			"ignoring.\n");
-		return;
-
+#ifdef WCM_CUSTOM_DEBUG
+	/* detect tool change */
+	if (lastPenSerial != common->wcmLastToolSerial) 
+	{	   
+		DBG(2, common->debugLevel, ErrorF("%s - usbParse: oldPen %d, newPen %d\n", 
+			timestr(), lastPenSerial, common->wcmLastToolSerial));
+		lastPenSerial = common->wcmLastToolSerial;
 	}
+#endif
+	channel = usbChooseChannel(local);
 
-	usbParseChannel(local,channel,common->wcmLastToolSerial);
-	common->wcmLastToolSerial = 0;
+	/* couldn't decide channel? invalid data */
+	if (channel == -1) goto skipEvent;
+
+	if (!common->wcmChannel[channel].work.proximity)
+	{
+		memset(&common->wcmChannel[channel],0,sizeof(WacomChannel));
+		/* in case the in-prox event was missing */
+		common->wcmChannel[channel].work.proximity = 1;
+
+#ifdef WCM_CUSTOM_DEBUG
+		detectChannelChange(local, channel);
+#endif
+ 	}
+	/* dispatch event */
+	usbParseChannel(local, channel);
+
+skipEvent:
 	common->wcmEventCnt = 0;
 }
 
-static void usbParseChannel(LocalDevicePtr local, int channel, int serial)
+static struct
+{
+	unsigned long device_type;
+	unsigned long tool_key;
+} wcmTypeToKey [] =
+{
+	{ STYLUS_ID, BTN_TOOL_PEN       },
+	{ STYLUS_ID, BTN_TOOL_PENCIL    },
+	{ STYLUS_ID, BTN_TOOL_BRUSH     },
+	{ STYLUS_ID, BTN_TOOL_AIRBRUSH  },
+	{ ERASER_ID, BTN_TOOL_RUBBER    },
+	{ CURSOR_ID, BTN_TOOL_MOUSE     },
+	{ CURSOR_ID, BTN_TOOL_LENS      },
+	{ TOUCH_ID,  BTN_TOOL_DOUBLETAP },
+	{ TOUCH_ID,  BTN_TOOL_TRIPLETAP },
+	{ PAD_ID,    BTN_TOOL_FINGER    }
+};
+
+static void usbParseChannel(LocalDevicePtr local, int channel)
 {
 	int i, shift, nkeys;
 	WacomDeviceState* ds;
 	struct input_event* event;
 	WacomDevicePtr priv = (WacomDevicePtr)local->private;
 	WacomCommonPtr common = priv->common;
+	WacomChannelPtr pChannel = common->wcmChannel + channel;
+	WacomDeviceState dslast = pChannel->valid.state;
+	static WacomDeviceState* syslast;
 
+	DBG(6, common->debugLevel, ErrorF("usbParseChannel %d events received\n", common->wcmEventCnt));
 	#define MOD_BUTTONS(bit, value) do { \
 		shift = 1<<bit; \
 		ds->buttons = (((value) != 0) ? \
 		(ds->buttons | (shift)) : (ds->buttons & ~(shift))); \
 		} while (0)
 
-	DBG(6, common->debugLevel, ErrorF("usbParseChannel \n"));
+	if (common->wcmEventCnt == 1 && !common->wcmEvents->type) {
+		DBG(6, common->debugLevel, ErrorF("usbParseChannel no real events received\n"));
+		return;
+	}
+
 	/* all USB data operates from previous context except relative values*/
 	ds = &common->wcmChannel[channel].work;
 	ds->relwheel = 0;
-	ds->serial_num = serial;
+	ds->serial_num = common->wcmLastToolSerial;
 
 	/* loop through all events in group */
 	for (i=0; i<common->wcmEventCnt; ++i)
@@ -856,22 +1032,34 @@ static void usbParseChannel(LocalDevicePtr local, int channel, int serial)
 				ds->stripx = event->value; 
 			else if (event->code == ABS_RY)
 				ds->stripy = event->value;
-			else if (event->code == ABS_RZ) 
-				ds->rotation = event->value;
-			else if (event->code == ABS_TILT_X)
+			else if (event->code == ABS_RZ) {
+				ds->rotation = event->value - MIN_ROTATION;
+				ds->rotation *= FILTER_PRESSURE_RES;
+				ds->rotation /= (MAX_ROTATION - MIN_ROTATION);
+			} else if (event->code == ABS_TILT_X)
 				ds->tiltx = event->value - common->wcmMaxtiltX/2;
 			else if (event->code ==  ABS_TILT_Y)
 				ds->tilty = event->value - common->wcmMaxtiltY/2;
-			else if (event->code == ABS_PRESSURE)
+			else if (event->code == ABS_PRESSURE) {
 				ds->pressure = event->value;
-			else if (event->code == ABS_DISTANCE)
+#ifdef WCM_CUSTOM_DEBUG
+				detectPressureIssue(event, common, channel);
+#endif
+
+			} else if (event->code == ABS_DISTANCE)
 				ds->distance = event->value;
-			else if (event->code == ABS_WHEEL || 
-				    event->code == ABS_Z)
-				ds->abswheel = event->value;
-			else if (event->code == ABS_THROTTLE)
-				ds->throttle = event->value;
-			else if (event->code == ABS_MISC && event->value)
+			else if (event->code == ABS_WHEEL) {
+				ds->abswheel = event->value * FILTER_PRESSURE_RES;
+				ds->abswheel /= MAX_ABS_WHEEL;
+			} else if (event->code == ABS_Z) {
+				ds->abswheel = event->value - MIN_ROTATION;
+				ds->abswheel *= FILTER_PRESSURE_RES;
+				ds->abswheel /= (MAX_ROTATION - MIN_ROTATION);
+			} else if (event->code == ABS_THROTTLE) {
+				ds->throttle = event->value + MAX_ABS_WHEEL;
+				ds->throttle *= FILTER_PRESSURE_RES;
+				ds->throttle /= (2 * MAX_ABS_WHEEL);
+			} else if (event->code == ABS_MISC && event->value)
 				ds->device_id = event->value;
 		}
 		else if (event->type == EV_REL)
@@ -890,6 +1078,9 @@ static void usbParseChannel(LocalDevicePtr local, int channel, int serial)
 				(event->code == BTN_TOOL_AIRBRUSH))
 			{
 				ds->device_type = STYLUS_ID;
+				/* V5 tools use ABS_MISC to report device_id */
+				if (common->wcmProtocolLevel == 4)
+					ds->device_id = STYLUS_DEVICE_ID;
 				ds->proximity = (event->value != 0);
 				DBG(6, common->debugLevel, ErrorF(
 					"USB stylus detected %x\n",
@@ -898,39 +1089,70 @@ static void usbParseChannel(LocalDevicePtr local, int channel, int serial)
 			else if (event->code == BTN_TOOL_RUBBER)
 			{
 				ds->device_type = ERASER_ID;
+				/* V5 tools use ABS_MISC to report device_id */
+				if (common->wcmProtocolLevel == 4)
+					ds->device_id = ERASER_DEVICE_ID;
 				ds->proximity = (event->value != 0);
 				if (ds->proximity)
 					ds->proximity = ERASER_PROX;
 				DBG(6, common->debugLevel, ErrorF(
-					"USB eraser detected %x\n",
-					event->code));
+					"USB eraser detected %x (value=%d)\n",
+					event->code, event->value));
 			}
 			else if ((event->code == BTN_TOOL_MOUSE) ||
 				(event->code == BTN_TOOL_LENS))
 			{
 				DBG(6, common->debugLevel, ErrorF(
-					"USB mouse detected %x\n",
-					event->code));
+					"USB mouse detected %x (value=%d)\n",
+					event->code, event->value));
 				ds->device_type = CURSOR_ID;
+				/* V5 tools use ABS_MISC to report device_id */
+				if (common->wcmProtocolLevel == 4)
+					ds->device_id = CURSOR_DEVICE_ID;
 				ds->proximity = (event->value != 0);
 			}
 			else if (event->code == BTN_TOOL_FINGER)
 			{
 				DBG(6, common->debugLevel, ErrorF(
-					"USB Pad detected %x\n",
-					event->code));
+					"USB Pad detected %x (value=%d)\n",
+					event->code, event->value));
 				ds->device_type = PAD_ID;
+				ds->device_id = PAD_DEVICE_ID;
 				ds->proximity = (event->value != 0);
 			}
-			else if (event->code == BTN_TOUCH)
+			else if (event->code == BTN_TOOL_DOUBLETAP)
 			{
-			/* reserved for touch now. Ping March 19, 2008
 				DBG(6, common->debugLevel, ErrorF(
-					"USB Touch detected %x\n",
-					event->code));
+					"USB Touch detected %x (value=%d)\n",
+					event->code, event->value));
 				ds->device_type = TOUCH_ID;
+				ds->device_id = TOUCH_DEVICE_ID;
 				ds->proximity = event->value;
-			*/
+				/* time stamp for 2GT gesture events */
+				if ((ds->proximity && !dslast.proximity) ||
+					(!ds->proximity && dslast.proximity))
+					ds->sample = GetTimeInMillis();
+
+				/* Bamboo touch tool doesn't send left button now */
+				if (!((common->tablet_id >= 0xD0) &&
+				    (common->tablet_id <= 0xD3) && 
+				    (ds->device_type == TOUCH_ID))) 
+					MOD_BUTTONS (0, event->value);
+			}
+			else if (event->code == BTN_TOOL_TRIPLETAP)
+			{
+				DBG(6, common->debugLevel, ErrorF(
+					"USB Touch second finger detected %x (value=%d)\n",
+					event->code, event->value));
+				ds->device_type = TOUCH_ID;
+				ds->device_id = TOUCH_DEVICE_ID;
+				ds->proximity = event->value;
+				/* time stamp for 2GT gesture events */
+				if ((ds->proximity && !dslast.proximity) ||
+					    (!ds->proximity && dslast.proximity))
+					ds->sample = (int)GetTimeInMillis();
+				/* Second finger events will be considered in 
+				 * combination with the first finger data */
 			}
 			else if ((event->code == BTN_STYLUS) ||
 				(event->code == BTN_MIDDLE))
@@ -950,22 +1172,60 @@ static void usbParseChannel(LocalDevicePtr local, int channel, int serial)
 				MOD_BUTTONS (4, event->value);
 			else
 			{
-				for (nkeys = 0; nkeys < common->npadkeys; nkeys++)
-					if (event->code == common->padkey_code [nkeys])
+				/* go through the whole array since usbRetrieveKeys 
+				 * may not get all keys on older kernels */
+				for (nkeys = 0; nkeys < sizeof (padkey_codes) / 
+						sizeof (padkey_codes [0]); nkeys++)
+					if (event->code == padkey_codes [nkeys])
 					{
-						MOD_BUTTONS ((MAX_MOUSE_BUTTONS/2+nkeys), event->value);
+						MOD_BUTTONS (nkeys, event->value);
 						break;
 					}
 			}
+
+#ifdef WCM_CUSTOM_DEBUG
+			/* detect prox out events */
+			if (dslast.proximity && ds->proximity == 0) {
+				DBG(2, common->debugLevel, ErrorF("%s - usbParseChannel: prox out"
+					" for %d, channel %d\n", timestr(), ds->serial_num, channel));
+                                DBG(3, common->debugLevel, dumpEventRing(local));
+			}
+#endif
 		}
 	} /* next event */
 
-	/* it is an out-prox when id or/and serial number is zero for kernel 2.4 */
-	if ((!ds->device_id || !ds->serial_num) && !USE_SYN_REPORTS(common))
- 		ds->proximity = 0;
+	/* First time in prox and device type unknown?
+	 * Tool may be on the tablet when X starts.
+	 */
+	if (!ds->device_type && !dslast.proximity)
+	{
+		usbRetrieveKeys(common);
+		for (i=0; i<sizeof(wcmTypeToKey) / sizeof(wcmTypeToKey[0]); i++)
+		{
+			if (ISBITSET(common->wcmKeys, wcmTypeToKey[i].tool_key))
+			{
+				ds->device_type = wcmTypeToKey[i].device_type;
+				break;
+			}
+		}
+	}
 
-	/* DTF720 doesn't support eraser */
-	if (common->tablet_id == 0xC0 && ds->device_type == ERASER_ID) 
+	/* don't send touch event when touch isn't enabled */
+	if ((ds->device_type == TOUCH_ID) && !common->wcmTouch)
+		return;
+
+	/* don't send touch event when pen is in prox */
+	if (syslast && ((ds->device_type == TOUCH_ID) &&
+		    ((syslast->device_type == STYLUS_ID) ||
+		    (syslast->device_type == ERASER_ID)) &&
+		    syslast->proximity))
+		return;
+
+	syslast = &common->wcmChannel[channel].work;
+
+	/* DTF720 and DTF720a don't support eraser */
+	if (((common->tablet_id == 0xC0) || (common->tablet_id == 0xC2)) && 
+			(ds->device_type == ERASER_ID)) 
 	{
 		DBG(10, common->debugLevel, ErrorF("usbParseChannel "
 			"DTF 720 doesn't support eraser "));
